import { ValidatorFn } from '@angular/forms';
import { FormControls, FormControlState, FormControlSummary, FormControlUpdate, FormGroupControlStates, FormGroupControlSummaries, FormGroupControlUpdates, Validator, FormGroupErrors, FormArrayErrors, FormControlErrors } from './types';
/**
 * Maps an object of control states to a value given by the `mapFunc`.
 *
 * @param controls Object which contains keys and associated control update.
 * @param mapFunc Mapper function to convert the update.
 *
 * @example
 * // Mapping of:
 * const controls = {
 *  firstControl: {
 *    value: 'firstValue',
 *    ...
 *  },
 *  secondControl: {
 *    value: 'secondValue',
 *    ...
 *  },
 * }
 *
 * // With the function:
 * const mapFunc = (control) => control.value
 *
 * // Will result in:
 * const result = {
 *  firstControl: 'firstValue',
 *  secondControl: 'secondValue'
 * }
 */
export declare function mapFormGroupControlStates<TControls extends FormControls, R>(controls: FormGroupControlStates<TControls>, mapFunc: (control: FormControlState<any>, key: string) => R): {
    [K in keyof TControls]: R;
};
/**
 * Maps an object of control summaries to a value given by the `mapFunc`.
 *
 * @param controls Object which contains keys and associated control summaries.
 * @param mapFunc Mapper function to convert the update.
 *
 * @example
 * // Mapping of:
 * const controls = {
 *  firstControl: {
 *    value: 'firstValue',
 *    ...
 *  },
 *  secondControl: {
 *    value: 'secondValue',
 *    ...
 *  },
 * }
 *
 * // With the function:
 * const mapFunc = (control) => control.value
 *
 * // Will result in:
 * const result = {
 *  firstControl: 'firstValue',
 *  secondControl: 'secondValue'
 * }
 */
export declare function mapFormGroupControlSummaries<TControls extends FormControls, R>(controls: FormGroupControlSummaries<TControls>, mapFunc: (control: FormControlSummary<any>, key: string) => R): {
    [K in keyof TControls]: R;
};
/**
 * Maps an object of control updates to a value given by the `mapFunc`.
 *
 * @param controls Object which contains keys and associated control updates.
 * @param mapFunc Mapper function to convert the update.
 *
 * @example
 * // Mapping of:
 * const controls = {
 *  firstControl: {
 *    value: 'firstValue',
 *    ...
 *  },
 *  secondControl: {
 *    value: 'secondValue',
 *    ...
 *  },
 * }
 *
 * // With the function:
 * const mapFunc = (control) => control.value
 *
 * // Will result in:
 * const result = {
 *  firstControl: 'firstValue',
 *  secondControl: 'secondValue'
 * }
 */
export declare function mapFormGroupControlUpdates<TControls extends FormControls, R>(controls: FormGroupControlUpdates<TControls>, mapFunc: (control: FormControlUpdate<any>, key: string) => R): {
    [K in keyof TControls]: R;
};
export declare function mapFormControls<TControls extends FormControls, R>(controls: {}, mapFunc: (control: {}, key: string) => R): {
    [K in keyof TControls]: R;
};
/**
 * Converts an Angular validator for the usage with this framework.
 * Does **not** support asynchronous validators.
 *
 * Only a subset of properties is available for the Angular validators:
 * `value, dirty, pristine, touched, untouched, enabled, disabled`
 * @param fn Angular validator function..
 */
export declare function validatorOf<T>(fn: ValidatorFn): Validator<T>;
/**
 * Merges an array of `FormControlErrors`.
 * Returns null if all errors are null.
 * @param errors An array of `FormControlErrors` which will be merged into a single `FormControlErrors` value.
 */
export declare function mergeFormControlErrors(...errors: FormControlErrors[]): FormControlErrors | null;
/**
 * Merges an array of errors.
 * The different child control errors will also be merged.
 * Returns null if all errors are null.
 * @param errors Multiple errors to be merged.
 *
 * @example
 * // Merge of:
 * {
 *   stringControl: {
 *     firstError: 'firstError'
 *   }
 * },
 * {
 *   stringControl: {
 *     secondError: 'secondError'
 *   }
 * }
 *
 * // Will result in:
 * {
 *   stringControl: {
 *     firstError: 'firstError',
 *     secondError: 'secondError'
 *   }
 * }
 *
 */
export declare function mergeFormGroupErrors<TControls extends FormControls>(...errors: FormGroupErrors<TControls>[]): FormGroupErrors<TControls> | null;
/**
 * Merges an array of errors.
 * Errors at each index will also be merged.
 * Returns null if all errors are null.
 * @param errors Multiple errors to be merged.
 *
 * @example
 * // Merge of:
 * [
 *   {
 *     firstError: 'firstError'
 *   }
 * ],
 * [
 *   stringControl: {
 *     secondError: 'secondError'
 *   }
 * ]
 *
 * // Will result in:
 * [
 *   {
 *     firstError: 'firstError',
 *     secondError: 'secondError'
 *   }
 * ]
 *
 */
export declare function mergeFormArrayErrors(...errors: FormArrayErrors[]): FormArrayErrors | null;
